from __future__ import absolute_import, division, print_function, unicode_literals

STARS = {
	"Sun": {
		"mass": 1.9891e30,
	}
}

PLANETS = {
	"Earth": {
		"mass": 5.9736e24,
	}
}

UNITS = {
	"Astronomical_unit": 149597870700, # m
	"Day": 60*60*24, # s
}

FUNDAMENTAL = {
	"Gravitational_constant": 6.67384e-11,
}
